////////////////////////////////////////////////////////////////////////
//
// SimWireT962 class designed to simulate signal on a wire in the TPC
//
// katori@fnal.gov
//
// - Revised to use sim::RawDigit instead of rawdata::RawDigit, and to
// - save the electron clusters associated with each digit.
// 
// wwu@fnal.gov
//
// - Add data driven noise model
//
////////////////////////////////////////////////////////////////////////

extern "C" {
#include <sys/types.h>
#include <sys/stat.h>
}

// LArSoft includes
#include "larcore/Geometry/Geometry.h"
#include "larcorealg/Geometry/CryostatGeo.h"
#include "larcorealg/Geometry/TPCGeo.h"
#include "larcorealg/Geometry/PlaneGeo.h"
#include "lardataobj/Simulation/sim.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RawData/raw.h"
#include "lardata/Utilities/LArFFT.h"
#include "lardata/DetectorInfoServices/DetectorClocksService.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"

// ROOT includes
#include <TMath.h>
#include <TH1D.h>
#include <TFile.h>
#include <TF1.h>
#include "TComplex.h"

// C++ includes
#include <vector>
#include <string>
#include <algorithm>
#include <sstream>
#include <fstream>
#include <bitset>
#include <random>

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDProducer.h"
#include "art/Framework/Principal/Event.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art_root_io/TFileService.h"
#include "art_root_io/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "cetlib/search_path.h"
#include "cetlib_except/exception.h"

// art extensions
#include "nurandom/RandomUtils/NuRandomService.h"

#include "CLHEP/Random/RandFlat.h"
#include "CLHEP/Random/RandGaussQ.h"

// ArgoNeuT includes
#include "T962Utilities/SignalShapingServiceT962.h"


namespace geo { class Geometry; }

///Detector simulation of raw signals on wires
namespace detsim {

  // Base class for creation of raw signals on wires.
  class SimWireT962 : public art::EDProducer {
  public:
    explicit SimWireT962(fhicl::ParameterSet const& pset);

  private:
    void produce (art::Event& evt) override;
    void beginJob() override;

    std::string            fDriftEModuleLabel;///< module making the ionization electrons
    raw::Compress_t        fCompression{raw::kNone}; ///< compression type to use
    double                 fNoiseFact;        ///< noise scale factor
    double                 fNoiseWidth;       ///< exponential noise width (kHz)
    double                 fLowCutoff;        ///< low frequency filter cutoff (kHz)
    bool                   fUseCoherentFeature; // if false, for ddn, simulated noise agrees with ddn
    bool                   fUseDataDrivenNoiseModel;
    std::vector<double>    fDDNFunctionParametersPlane0; // data driven noise function parameters for plane 0
    std::vector<double>    fDDNFunctionParametersPlane1; // data driven noise function parameters for plane 1
    int                    fSimulationIndex; // 0: only noise; 1: only signal; 2: noise + signal

    int                    fNTicks;           ///< number of ticks of the clock
    double  fSampleRate;       ///< sampling rate in ns
    std::vector<std::vector<float>> fNoise;///< noise on each channel for each time
    void  GenNoise(std::vector<float>& array, unsigned int channel); // noise model for plane channels
    TH1D* fNoiseDist;        ///< distribution of noise counts
    CLHEP::HepRandomEngine& fEngine; ///< reference to art-managed random-number engine

  }; // class SimWireT962

}

namespace detsim{

  //-------------------------------------------------
  SimWireT962::SimWireT962(fhicl::ParameterSet const& pset)
    : EDProducer{pset}
    , fDriftEModuleLabel{pset.get<std::string>("DriftEModuleLabel")}
    , fNoiseFact        {pset.get<double>("NoiseFact")}
    , fNoiseWidth       {pset.get<double>("NoiseWidth")}
    , fLowCutoff        {pset.get<double>("LowCutoff")}
    , fUseCoherentFeature   {pset.get<bool>("UseCoherentFeature")}
    , fUseDataDrivenNoiseModel   {pset.get<bool>("UseDataDrivenNoiseModel")}
    , fDDNFunctionParametersPlane0  {pset.get<std::vector<double>>("DDNFunctionParametersPlane0")}
    , fDDNFunctionParametersPlane1  {pset.get<std::vector<double>>("DDNFunctionParametersPlane1")}
    , fSimulationIndex  {pset.get<int>("SimulationIndex", 2)}
    // create a default random engine; obtain the random seed from NuRandomService,
    // unless overridden in configuration with key "Seed"
    , fEngine(art::ServiceHandle<rndm::NuRandomService>()->createEngine(*this, pset, "Seed"))
  {
    produces< std::vector<raw::RawDigit>   >();

    std::string compression(pset.get< std::string >("CompressionType"));
    if(compression.compare("Huffman") == 0) fCompression = raw::kHuffman;
  }

  //-------------------------------------------------
  void SimWireT962::beginJob()
  {
    auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService const>()->DataForJob();
    auto const detProp = art::ServiceHandle<detinfo::DetectorPropertiesService const>()->DataForJob(clockData);
    fSampleRate = sampling_rate(clockData);

    art::ServiceHandle<util::LArFFT> fFFT;
    //fFFT->ReinitializeFFT(fNTimeSamples,fFFT->FFTOptions(),fFFT->FFTFitBins());
    // www: default FFTSize for wire simulation is 4096, which is twice of the readout window (2048);
    // data driven noise model is from fitting to noise in readout window (2048)
    if (fUseDataDrivenNoiseModel) fNTicks = detProp.NumberTimeSamples(); // 2048
    else  fNTicks = fFFT->FFTSize();

    // get access to the TFile service
    art::ServiceHandle<art::TFileService> tfs;
    fNoiseDist = tfs->make<TH1D>("Noise", ";Noise (ADC);", 1000, -10., 10.);
    art::ServiceHandle<geo::Geometry> geo;
    fNoise.resize(geo->Nchannels());
    // GenNoise() will further resize each channel's
    // noise vector to fNTicks long.
    for(unsigned int p = 0; p < geo->Nchannels(); ++p){
      GenNoise(fNoise[p], p);
      for(int i = 0; i < fNTicks; ++i) fNoiseDist->Fill(fNoise[p][i]);
    }// end loop over wires
  } // beginJob

  //-------------------------------------------------
  void SimWireT962::produce(art::Event& evt)
  {

    // get the geometry to be able to figure out signal types and chan -> plane mappings
    art::ServiceHandle<geo::Geometry> geo;
    auto const clockData = art::ServiceHandle<detinfo::DetectorClocksService const>()->DataFor(evt);
    unsigned int signalSize = fNTicks/2;

    // generator noise for each event: this is necessary for noise study, i.e., cnn roi
    if (fUseDataDrivenNoiseModel) {
      fNoise.clear();
      fNoise.resize(geo->Nchannels());
      for(unsigned int p = 0; p < geo->Nchannels(); ++p){
        GenNoise(fNoise[p], p);
      }// end loop over wires
    }

    // www: wire sim uses a wide window (4096 ticks); data driven noise is from a readout window of 2048 ticks
    art::ServiceHandle<util::LArFFT> fFFT;
    if (fUseDataDrivenNoiseModel) {
      fFFT->ReinitializeFFT(fNTicks*2,fFFT->FFTOptions(),fFFT->FFTFitBins()); //can affect Convolute
      signalSize = fNTicks;
    }

    // vectors for working
    std::vector<short> adcvec(signalSize, 0);
    std::vector<double> chargeWork(fNTicks);
    if (fUseDataDrivenNoiseModel) chargeWork.resize(fNTicks*2);

    std::vector<const sim::SimChannel*> chanHandle;
    evt.getView(fDriftEModuleLabel,chanHandle);

    // make a vector of const sim::SimChannel* that has same number
    // of entries as the number of channels in the detector
    // and set the entries for the channels that have signal on them
    // using the chanHandle
    std::vector<const sim::SimChannel*> channels(geo->Nchannels());
    for(size_t c = 0; c < chanHandle.size(); ++c) channels[chanHandle[c]->Channel()] = chanHandle[c];

    // make an unique_ptr of sim::SimDigits that allows ownership of the produced
    // digits to be transferred to the art::Event after the put statement below
    std::unique_ptr<std::vector<raw::RawDigit>>  digcol(new std::vector<raw::RawDigit>);

    art::ServiceHandle<util::SignalShapingServiceT962> sss;

    // Add all channels
    CLHEP::RandFlat flat(fEngine);

    std::map<int,double>::iterator mapIter;
    float pedrms = 0;
    float pedestal = 400;
    for(unsigned int chan = 0; chan < geo->Nchannels(); chan++) {

      std::fill(chargeWork.begin(), chargeWork.end(), 0.);

      if( channels[chan] ){
        // get the sim::SimChannel for this channel
        const sim::SimChannel* sc = channels[chan];

        // loop over the tdcs and grab the number of electrons for each
        for(size_t t = 0; t < chargeWork.size(); ++t){
          int tdc = clockData.TPCTick2TDC(t);
          if (tdc < 0) continue;
          if (tdc >= int(chargeWork.size())) continue;
          chargeWork[t] = sc->Charge(tdc);
        } // t
        sss->Convolute(chan, chargeWork);
      } // channels[chan]
/*
      if(chan == 103) {
        unsigned short lobin = 340;
        for(unsigned short bin = lobin; bin < lobin+100; ++bin) mf::LogVerbatim("SignalShapingServiceT962")<<chargeWork[bin];
      }
*/
      // noise was already generated for each wire in the event
      // raw digit vec is already in channel order
      // pick a new "noise channel" for every channel  - this makes sure
      // the noise has the right coherent characteristics to be on one channel
      // www: one should pay attention to the simulated window while combining noise and signal.
      int noisechan = TMath::Nint(flat.fire()*(1.*(fNoise.size()-1)+0.1));
      for(unsigned int i = 0; i < signalSize; ++i) {
        if (fUseCoherentFeature) adcvec[i] = (short)TMath::Nint(fNoise[noisechan][i] + chargeWork[i+60] + pedestal);
        //else adcvec[i] = (short)TMath::Nint(fNoise[chan][i] + chargeWork[i] + pedestal); // for simultion, "short" can affect the noise frequency a little
        else {
          if (fSimulationIndex == 2) adcvec[i] = (short)TMath::Nint(fNoise[chan][i] + chargeWork[i] + pedestal); // "short" can affect the noise frequency a little; save noise + signal
          else if (fSimulationIndex == 1) adcvec[i] = (short)TMath::Nint(chargeWork[i] + pedestal); // only save signal
          else if (fSimulationIndex == 0) adcvec[i] = (short)TMath::Nint(fNoise[chan][i] + pedestal); // only save noise
        }
      } // i

      // compress the adc vector using the desired compression scheme,
      // if raw::kNone is selected nothing happens to adcvec
      // This shrinks adcvec, if fCompression is not kNone.
      raw::Compress(adcvec, fCompression);

      raw::RawDigit rd(chan, signalSize, adcvec, fCompression);
      rd.SetPedestal(pedestal, pedrms);
      // Then, resize adcvec back to full length!
      adcvec.clear();
      adcvec.resize(signalSize,0.0);
      // add this digit to the collection
      digcol->push_back(rd);
    }// end loop over channels

    evt.put(std::move(digcol));
  } // produce

  //-------------------------------------------------
  void SimWireT962::GenNoise(std::vector<float>& noise, unsigned int channel)
  {
    CLHEP::RandFlat flat(fEngine);

    noise.resize(fNTicks);
    std::fill(noise.begin(), noise.end(), 0.);
    std::vector<TComplex> noiseFrequency(fNTicks, 0.);///<noise in frequency space

    double pval = 0.;
    double lofilter = 0.;
    double phase = 0.;
    double rnd[2] = {0.};

    // width of frequencyBin in kHz: for data driven noise is 2.46 kHz
    double binWidth = 1.0/(fNTicks*fSampleRate*1.0e-6);

    // gain function (data driven noise model frequncy magnitude mean): fit with modified microboone noise model after coherent noise removal
    TF1* ddn_model_f1 = new TF1("ddn_model_f1", "[0]*exp(-0.5*pow((x-[1])/[2],2))*([3]/(x+[4])+1) + [5] + exp(-[6]*(x-[7]))", 0., fNTicks*binWidth);

    double ddn_par[8];
    for (int ipar=0; ipar<8; ipar++) {
       if (channel < 240) ddn_par[ipar] = fDDNFunctionParametersPlane0[ipar];
       else ddn_par[ipar] = fDDNFunctionParametersPlane1[ipar];
    }

    ddn_model_f1->SetParameters(ddn_par);
    
    // data driven noise model: frequency magnitude fluctuation: weighted Poisson distribution
    TF1* f1_magnitude_model = new TF1("f1_magnitude_model", "[0]*pow([1]/[2], x/[2])*exp(-[1]/[2])/ROOT::Math::tgamma(x/[2]+1.)", 0, 5.);
    f1_magnitude_model->SetParameter(0, 0.7025);
    f1_magnitude_model->SetParameter(1, 0.9914);
    f1_magnitude_model->SetParameter(2, 0.2865);
    // www: from data draven noise: freq inverse FFT gives data driven noise.
    // raw::Compress treats adcvec(noise) as a "short" type, instead of double. The frequency from  FFT of the simulated noise will be affected by the "short" type slightly.
    double scalefactor_a = 1.59; // for very low freq (<2.46kHz)
    double scalefactor_b = 0.77; // for low freq (2.46kHz, 4.92kHz)

    for(int i=0; i< fNTicks; ++i){
      if (fUseDataDrivenNoiseModel) {
        // data driven noise
        if (i<fNTicks/2) pval =  ddn_model_f1->Eval((i+0.5)*binWidth);
        else pval =  ddn_model_f1->Eval((fNTicks-i-0.5)*binWidth);

        if (i==0 || i==fNTicks-1) pval *= scalefactor_a;
        if (i==1 || i==fNTicks-2) pval *= scalefactor_b;

        // noise frequency magnitiude model
        double randomizer = f1_magnitude_model->GetRandom();
        pval = pval*randomizer/fNTicks;

        phase = gRandom->Uniform(0,1) * 2. * TMath::Pi();
      }
      else {
        // exponential noise spectrum
        pval = fNoiseFact*exp(-(double)i*binWidth/fNoiseWidth);
        //low frequency cutoff
        lofilter = 1.0/(1.0+exp(-(i-fLowCutoff/binWidth)/0.5));
        //randomize 10%
        flat.fireArray(2,rnd,0,1);
        pval *= lofilter*(0.9+0.2*rnd[0]);
        // random pahse angle
        phase = rnd[1]*2.*TMath::Pi();
      }

      TComplex tc(pval*cos(phase),pval*sin(phase));
      noiseFrequency[i] += tc;
    }

    // inverse FFT MCSignal
    art::ServiceHandle<util::LArFFT> fFFT;
    // change the FFTSize according to fNTicks
    fFFT->ReinitializeFFT(fNTicks,fFFT->FFTOptions(),fFFT->FFTFitBins());
    fFFT->DoInvFFT(noiseFrequency, noise);

    noiseFrequency.clear();

    // multiply each noise value by fNTicks as the InvFFT
    // divides each bin by fNTicks assuming that a forward FFT
    // has already been done.
    for(unsigned int i = 0; i < noise.size(); ++i) noise[i] *= 1.*fNTicks;

  } // GenNoise

}

DEFINE_ART_MODULE(detsim::SimWireT962)
